import json

import pytest
from fastmcp import Client
from google.cloud.bigquery.enums import QueryApiMethod
from typer.testing import CliRunner

from bigquery_mcp.main import mcp_app, MCPProtocol, cli_app


@pytest.fixture(scope="module")
def app():
    return mcp_app(
        mode=MCPProtocol.studio,
        dataset=["bigquery-public-data.usa_names"],
        table=[],
        project=None,
        api_method=QueryApiMethod.QUERY,
        port=8000,
    )


# noinspection SqlNoDataSourceInspection,SqlDialectInspection
@pytest.fixture
def public_query():
    return """
           SELECT
               name,
               SUM(number) AS total
           FROM
               `bigquery-public-data.usa_names.usa_1910_2013`
           GROUP BY
               name
           ORDER BY
               total DESC
               LIMIT
               2;
           """


async def test_cli(app):
    runner = CliRunner()
    result = runner.invoke(cli_app, ["--help"])

    assert result.exit_code == 0
    assert "Usage:" in result.stdout
    assert "mode" in result.stdout


async def test_list_resources(app):
    async with Client(app) as client:
        resources = await client.list_resources()
        resource_templates = await client.list_resource_templates()
        assert len(resources) == 1
        assert len(resource_templates) == 1


async def test_get_resources(app):
    async with Client(app) as client:
        resources = await client.list_resources()
        resource_parts = await client.read_resource(resources[0].uri)
        resp = response_obj(resource_parts)
        assert len(resp) == 2
        assert "bigquery-public-data.usa_names.usa_1910_current" in resp


async def test_mcp_server_has_query_tool(app):
    async with Client(app) as client:
        tools = {t.name: t for t in await client.list_tools()}
        assert "query" in tools
        assert "sql" in tools["query"].inputSchema["properties"]


async def test_can_query(app, public_query):
    async with Client(app) as client:
        responses = await client.call_tool("query", dict(sql=public_query))
        rows = response_obj(responses)
        assert len(rows) == 2


async def test_can_get_schema(app, public_query):
    async with Client(app) as client:
        responses = await client.call_tool(
            "get_schema", dict(table="bigquery-public-data.usa_names.usa_1910_current")
        )
        schema = response_obj(responses)
        assert schema


async def test_bad_query_errors(app):
    async with Client(app) as client:
        responses = await client.call_tool("query", dict(sql="foo"))
        assert response_obj(responses)["error"] == dict(
            reason="invalidQuery",
            message='Syntax error: Unexpected identifier "foo" at [1:1]',
        )


def response_obj(responses) -> dict | list:
    rtn = [json.loads(r.text) for r in responses]
    return rtn[0] if len(rtn) == 1 else rtn



from __future__ import annotations

import logging

from fastmcp import FastMCP
from google.api_core.exceptions import BadRequest
from google.cloud.bigquery.table import RowIterator
from pydantic import Field

from bigquery_mcp.config import Config

logger = logging.getLogger(__name__)


def make_app(app: FastMCP, config: Config):
    @app.tool(
        description="Executes the provided BigQuery sql statement and returns the results"
    )
    def query(
        sql: str = Field(description="BigQuery sql statement to execute"),
    ):
        with config.get_client() as client:
            try:
                executed_query = client.query(sql, api_method=config.api_method)
                results: RowIterator = executed_query.result()
                rows = [dict(r.items()) for r in results]
                return rows
            except BadRequest as br:
                logger.warning(f"Bad request: {br}")
                fields = ["reason", "message"]
                errors = [{f: e[f] for f in fields if f in e} for e in br.errors]
                return dict(error=errors[0] if len(errors) == 1 else dict(error=errors))
            except Exception as e:
                logger.exception("Error executing query")
                return dict(error=str(e))

    @app.resource("bigquery://tables")
    def list_tables():
        with config.get_client() as client:
            tables: list[str] = [
                str(table.reference)
                for dataset in config.datasets
                for table in client.list_tables(dataset)
            ]
        for table in config.tables:
            if table.count(".") == 1:
                tables.append(config.project + "." + table if config.project else table)
            elif table.count(".") == 2:
                tables.append(table)
            else:
                logger.warning(f"Invalid table: {table}")
        return tables

    @app.resource(
        "bigquery://tables/{table}/schema",
        name="Table Schema",
        mime_type="application/json",
    )
    def get_schema(table: str) -> dict:
        with config.get_client() as client:
            table = client.get_table(table)
            table_dict = table.to_api_repr()
            desired_fields = [
                "tableReference",
                "description",
                "schema",
                "numBytes",
                "numRows",
                "creationTime",
                "lastModifiedTime",
                "resourceTags",
                "labels",
            ]
            return {
                field: table_dict[field]
                for field in desired_fields
                if field in table_dict
            }

    if config.enable_list_tables_tool:
        app.add_tool(list_tables, "list_tables", "List the tables available")
    if config.enable_schema_tool:
        app.add_tool(get_schema, "get_schema", "Get the schema for a given table")



from __future__ import annotations

from dataclasses import dataclass
from multiprocessing import Lock
from typing import List, Optional

from google.cloud import bigquery
from google.cloud.bigquery.enums import QueryApiMethod


@dataclass
class Config:
    datasets: List[str]
    tables: List[str]
    project: Optional[str]
    api_method: QueryApiMethod
    enable_schema_tool: bool = True
    enable_list_tables_tool: bool = True
    _client: bigquery.Client = None
    _lock: Lock = Lock()

    def get_client(self) -> bigquery.Client:
        if not self._client:
            with self._lock:
                if not self._client:
                    kwargs = {}
                    if self.project:
                        kwargs["project"] = self.project

                    self._client = bigquery.Client(**kwargs)
        return self._client



from enum import Enum
from functools import wraps
from typing import List, Optional

import typer
from fastmcp import FastMCP
from google.cloud.bigquery.enums import QueryApiMethod

from bigquery_mcp.config import Config
from bigquery_mcp.server import make_app

cli_app = typer.Typer(help="BigQuery MCP Server")


class MCPProtocol(str, Enum):
    studio = "stdio"
    sse = "sse"
    streamable_http = "streamable-http"


def mcp_app(
    mode: MCPProtocol = typer.Option(
        MCPProtocol.studio, help="MCP transport protocol"
    ),  # this is required for typer args
    dataset: List[str] = typer.Option(
        default=[],
        help="Dataset(s) for mcp resources. Will create resources for all tables.",
    ),
    table: List[str] = typer.Option(
        default=[],
        help="Table(s) for mcp resources. Can be specified as project.dataset.table or dataset.table",
    ),
    enable_list_tables_tool: bool = typer.Option(
        True, help="Register list_resources tool"
    ),
    enable_schema_tool: bool = typer.Option(True, help="Register get_schema tool"),
    project: Optional[str] = typer.Option(
        None, help="BigQuery project", envvar="BQ_PROJECT"
    ),
    api_method: QueryApiMethod = typer.Option(
        QueryApiMethod.QUERY, help="BigQuery client api_method"
    ),
    port: int = typer.Option(8000),
) -> FastMCP:
    config = Config(
        datasets=dataset,
        project=project,
        api_method=api_method,
        tables=table,
        enable_schema_tool=enable_schema_tool,
        enable_list_tables_tool=enable_list_tables_tool,
    )
    app = FastMCP("BigQuery MCP Server")
    make_app(app, config)
    return app


@cli_app.command()
@wraps(mcp_app)
def wrapper(**kwargs):
    app = mcp_app(**kwargs)
    run_args = dict(transport=kwargs["mode"])
    if kwargs["mode"] != "stdio":
        run_args["port"] = kwargs["port"]
    app.run(**run_args)


def main():
    cli_app()


if __name__ == "__main__":
    main()
